/* DC-IRC Test Bed - The Dining Cryptographers IRC Test Bed Client
   Alpha Version
   Copyright (C) 1994 Thomas G. Edwards 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   I encourage you to make changes and updates to this software,
   just email them to me when you are done!  Collisions, adaptive
   adding of new clients into the conversation, etc. need to be handled.

   Send your comments, updates to this software, and all your spare 
   pocket change to tedwards@src.umd.edu (Thomas G. Edwards).

   [Some of the IRC communication code is directly inspired from TinyIRC 
    Alpha Release by Nathan I. Laredo]

*/


#define IRCHOST "us.undernet.org"  /* irc server host to use */
/*#define IRCHOST "irc.bu.edu" */
#define IRCPORT 6667               /* irc port to connect to */

#define THRESH 10   /* number of seconds to send timeout */
#define WORDS 128  /* words exchanged each time */

#include<unistd.h>
#include<fcntl.h>
#include<stdio.h>
#include<sgtty.h>
#include<sys/ioctl.h>
#include<strings.h>
#include<sys/types.h>
#include<sys/time.h>
#include<sys/socket.h>
#include<netinet/in.h>
#include<netdb.h>
#include<signal.h>
#include<pwd.h>
#include<ctype.h>

/* globals */

int sockfd,        /* socket fd */
    cursd=0,       /* current column of incoming server info */
    sok=1;	   /* socket is ok */
    stdinfd,	   /* stdin file descriptor  */
    stdoutfd,	   /* stdout file descriptor */
    go=0, 	   /* session started or not */
    instigator=0,  /* whether client is instigator */
    ack_count,     /* count of ACKS */
    block_count,   /* count of DC data blocks */
    numdc,	   /* number of Dining Cryptographers */
    last_send,     /* time of last sending */
    send_OK,	   /* OK to start sending data? */
    got_flips, 	   /* got neighbors flips */
    flips_sent,    /* have we sent our flips already? */
    block_sent,    /* have we sent out block ? */
    ack_sent,      /* have we sent out an ack ? */
    msg_waiting;   /* a message is waiting to be transmitted */

int DC[WORDS];		           /* difference sum */
int my_flips[WORDS];		   /* your random flips */
int neighbor_flips[WORDS];	   /* neigbor random flips */
int same[WORDS];		   /* comparison between you and neighbor */

char lineout[512];    /* character string to be send to IRC server */
char localhost[64];   /* local host name */
char serverdata[512]; /* data coming from server */
char linein[512];     /* characters coming in */
char neighbor[20];    /* irc name of neighbor */
char ircname[20];     /* your irc name */
char msg_in[512];     /* message to communicate to dining cryptographers */

struct passwd *userinfo; /* passwd user information */

makeconnect()
/* connect to irc server */
{
    struct sockaddr_in sa;
    struct hostent *hp;
    int s;

    if ((hp = gethostbyname(IRCHOST)) == NULL)
        return -1;
    bzero(&sa, sizeof(sa));
    bcopy(hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
    sa.sin_family = hp->h_addrtype;
    sa.sin_port = htons((u_short) IRCPORT);
    if ((s = socket(hp->h_addrtype, SOCK_STREAM, 0)) < 0)
        return -1;
    if (connect(s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
        close(s);
        return -1;
    }
    fcntl(s, F_SETFL, O_NDELAY);
    return s;
}

int sendline()
/* send a line to the server */
{
    if (write(sockfd, lineout, strlen(lineout)) < 1)
        return 0;
    return 1;
}

void tell(msg)
char* msg;
/* make a public broadcast to the channel */
{
	sprintf(lineout,"PRIVMSG #dcnet :%s\n",msg);
	sendline();
}

void userinput()
/* process user input */
{
	if(go==0 && instigator==1)
		{
		fgets(linein, 500, stdin);
		tell("GO");
		printf("---Beginning DC Protocol---\n");	
		go=1;
		last_send=time(NULL);	
		return;
		}
	else
	if(go==0 && instigator==0)
		{
		fgets(linein, 500, stdin);
		tell(linein);
		}
	else
		{
		fgets(msg_in,500,stdin);
		msg_waiting=1;
		}
	
}

array_2_hex(array,hex)
int* array;
char* hex;
/* convert array to hex character string */
{
int i;

        for(i=0;i<(WORDS/2);++i)
        {
        sprintf(hex+(i*4),"%02x",array[i*2]);
        sprintf(hex+(i*4)+2,"%02x",array[(i*2)+1]);
        }
}

hex_2_array(array,hex)
int* array;
char* hex;
/* convert hex character string to array */
{
int i;

        for(i=0;i<(WORDS/2);++i)
                {
                sscanf(hex+(i*4),"%2x",&array[(i*2)]);
                sscanf(hex+(i*4)+2,"%2x",&array[(i*2)+1]);
                }
}

void process_block(data)
char *data;
/* get block of differences from broadcast channel and add to sum */
{
int new_data[WORDS];
int i;

	block_count++;
	hex_2_array(new_data,data);
	for(i=0;i<WORDS;++i)
		DC[i]=DC[i] ^ new_data[i];		
}

void process_flips(data)
char *data;
/* process neighbor flips */
{
	hex_2_array(neighbor_flips,data);
}

/* parsedata - take one line of input and process it */

int parsedata()
/* decide how to act on irc server messages */
{
char *tmp;
	
/* take care of ping immediately, lest irc kick you off */
	if(serverdata[0] == 'P')
		{
		sprintf(lineout, "PONG :%s\n", ircname);
		return sendline();
		}
/* strip off header */
	tmp=serverdata;
	if(serverdata[0]==':')
	{
		tmp++;
		while(tmp[0] != ':')
			tmp++;
		tmp++;
	}	

/* is it a GO command? */
	if(go==0 && (strncmp(tmp,"GO",2)==0))
	{
		printf("----GO received!----\n");
		go=1;
		last_send=time(NULL);	
		return 1;	
	}

/* echo and exit if not Going.. */

	if(go==0)
	{
	printf("%s\n",tmp);
	return 1;
	}

/* is it a data block? */

	if(strncmp(tmp,"DC",2)==0)
	{
	process_block(tmp+2);		 
	return 1;
	}

/* is it an ACK? */

	if(strncmp(tmp,"AK",2)==0)
	{
	ack_count++;
	return 1;
	}

/* is it a FLIP from a neighbor? */

	if(strncmp(tmp,"FL",2)==0)
	{
	process_flips(tmp+2);
	got_flips=1;
	return 1;
	}

	return 1;
}

int serverinput()
/* get input from server */
{
	char ich;
	while (read(sockfd, &ich, 1) > 0)
        if (ich != '\n' && ich != '\r')
            serverdata[cursd++] = ich;
        else {
            serverdata[cursd] = 0;
            if (cursd) {
                cursd = 0;
                return parsedata();
            }
        }
    return 1;
}

/* time to send? or is anyone else sending? */

time2send()
{
	if((time(NULL)-last_send>THRESH) ||
	   (block_count !=0))
		send_OK=1;
}

void ack()
/* acknowledge receipt of all differences */
{
	tell("AK");
	ack_count++;
	ack_sent=1;
}

void decrypt()
/* output characters based on sum of differences */
{
char tmpout[WORDS*2];
int i,anything=0;

	array_2_hex(DC,tmpout);
	for(i=0;i<WORDS;++i)
                if((DC[i]!=0))
                        {
			if(anything==0)
				printf("-->");	
			printf("%c",toascii(DC[i]));
			anything=1;
			}
	for(i=0;i<WORDS;++i)
		DC[i]=0;
	ack_count=0;
	block_count=0;
	send_OK=0;
	got_flips=0;
	block_sent=0;
	flips_sent=0;
	ack_sent=0;
	last_send=time(NULL);
}

void send_flips()
/* send your flips to neighbor */
{
long tmp;
char tmpout[512];
int i;
	
	for(i=0;i<(WORDS/2);++i)
		{
                tmp=random();
                my_flips[i*2]=(tmp & 0xFF00) >> 8;
                my_flips[(i*2)+1]=tmp & 0x00FF;
	       }
	array_2_hex(my_flips,tmpout);

/* send flips in private message - this could be more secure */

	sprintf(lineout, "PRIVMSG %s :FL%s\n",neighbor,tmpout);
	sendline();
	flips_sent=1;
}

void send_block()
/* send block of differences (modulated by message) to broadcast channel */
{
int i;
char tmpout[512],tmpout2[512];

	if(msg_waiting==1)
		printf("<Sending message>\n",msg_in);
	for(i=0;i<WORDS;++i)
		{
		same[i]=(my_flips[i] ^ neighbor_flips[i]);	
		if(msg_waiting==1)
			{
			same[i]=same[i] ^ msg_in[i]; /* modulate differences */
			msg_in[i]=0; /* erase old message as it is sent */
			}
		DC[i]=DC[i] ^ same[i]; /* include your own differences in sum */
		}
	array_2_hex(same,tmpout);	
	sprintf(tmpout2,"DC%s",tmpout);	
	tell(tmpout2);
	block_sent=1;
	if(msg_waiting==1)
		msg_waiting=0;
}

main()
{
fd_set readfs;		/* file set for select statement */
struct timeval timeout; /* time structure for timeout */
int rseed,		/* random seed */
	i;		

printf("Dining Cryptographers IRC Client\n");
printf("--------------------------------\n\n");
printf("Are you the instigator? (Y or N) ");
fgets(linein, 3, stdin);
if (linein[0]=='y' || linein[0]=='Y')
	instigator=1;
else
	instigator=0;

printf("Your irc name? ");
scanf("%s",ircname);

printf("How many Dining Cryptographers? ");
scanf("%d",&numdc);
printf("Who is your right-hand neighbor? ");
scanf("%s",neighbor);
printf("\nEnter random seed: ");
scanf("%d",&rseed);
srandom(rseed);

stdinfd = fileno(stdin);
stdoutfd = fileno(stdout);
userinfo = getpwuid(getuid());

printf("Making conntection...\n");
if ((sockfd = makeconnect()) < 0)
	{
	fprintf(stderr, "*** connection refused, aborting\n");
	exit(0);
	}
printf("connected to irc server!\n");
sprintf(lineout, "NICK :%s\n", ircname);
sendline();
printf("Sent NICK\n");
gethostname(localhost, 64);
sprintf(lineout, "USER %s %s %s :%s\n", userinfo->pw_name, localhost,
                IRCHOST, ircname);
sendline();
printf("User signed up\n");
sprintf(lineout,"JOIN #dcnet\n");
sendline();
printf("Joined #dcnet\n");

for(i=0;i<WORDS;++i)
	DC[i]=0;
block_count=0;
ack_count=0;
send_OK=0;
flips_sent=0;
block_sent=0;
ack_sent=0;
msg_waiting=0;

if (instigator)
	printf("\n\n-----Press return to start session--------\n\n");

while (sok) 
	{
	FD_ZERO(&readfs);
	FD_SET(sockfd, &readfs);
	FD_SET(stdinfd, &readfs);
	timeout.tv_sec = THRESH;
	timeout.tv_usec = 0;
	if (select(FD_SETSIZE, &readfs, NULL, NULL, &timeout)) 
        	{
	    	if (FD_ISSET(stdinfd, &readfs))
                	userinput();
            	if (FD_ISSET(sockfd, &readfs))
                	sok = serverinput();
		}
	if(go==1)
		{	
		if( (block_count==(numdc-1)) &&
				(block_sent==1) &&
				(flips_sent==1) &&
				(ack_sent==0))
			ack();
		if(ack_count==numdc)
			decrypt();
	 	time2send();
		if(send_OK && (flips_sent==0))		
			send_flips();
		if(got_flips && (block_sent==0))
			send_block();
		}	
	}
}
	
